import os
import numpy as np
import iris
import iris.cube
import iris.analysis
import warnings
import numba
import datetime
from numba import jit, njit
from numba import vectorize
from numba import guvectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver

MESH_FILE = "/esarchive/autosubmit/con_files/mesh_mask_nemo.Ec3.2_O1L75.nc"
TOS_FILE = "/esarchive/exp/ecearth/a13c/cmorfiles/CMIP/EC-Earth-Consortium"\
            "/EC-Earth3-LR/historical/r1i1p1f1/Omon/tos/gn/v20180501"\
            "/tos_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_195001-195012.nc"
THETAO_FILE = "/esarchive/exp/ecearth/a13c/cmorfiles/CMIP/EC-Earth-Consortium"\
              "/EC-Earth3-LR/historical/r1i1p1f1/Omon/thetao/gn/v20180501/"\
              "thetao_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_"\
              "195001-195012.nc"
REGIONS_FILE = "/esarchive/autosubmit/con_files/mask.regions.Ec3.2_O1L75.nc"


def main():
    e1t, e2t, e3t_0 = load_mesh()
    basins = load_masks()
    tos = load_tos()
    for i in range(1):
        start_gpu = datetime.datetime.now()
        regmean_tos_gpu = compute_regmean_2D_gpu(tos, basins, e1t, e2t)
        end_gpu = datetime.datetime.now() - start_gpu
        start_numpy = datetime.datetime.now()
        regmean_tos_numpy = compute_regmean_2D_numpy(tos, basins, e1t, e2t)
        end_numpy = datetime.datetime.now() - start_numpy
        print('Computing time 2D var GPU: {0} s'.format(end_gpu))
        print('Computing time 2D var numpy: {0} s'.format(end_numpy))

    del tos
    thetao = load_thetao()
    for i in range(1):
        start_gpu_3D = datetime.datetime.now()
        regmean_thetao_gpu = compute_regmean_gpu_3D(thetao, basins,  e1t, e2t,
                                                    e3t_0)
        regmean_thetao_lev_gpu = compute_regmean_lev_gpu_3D(thetao, basins,
                                                            e1t, e2t, e3t_0)
        end_gpu_3D = datetime.datetime.now() - start_gpu_3D
        start_numpy_3D = datetime.datetime.now()
        regmean_3D = compute_regmean_3D_numpy(thetao, basins, e1t, e2t, e3t_0)
        reg_lev = compute_regmean_3D_levels_numpy(thetao, e1t, e2t, e3t_0)
        end_numpy_3D = datetime.datetime.now() - start_numpy_3D
        print('Computing time 3D var GPU: {0} s'.format(end_gpu_3D))
        print('Computing time 3D var numpy: {0} s'.format(end_numpy_3D))


def load_mesh():
    e1t = iris.load_cube(MESH_FILE, 'e1t').data.astype(np.float32)
    e1t = np.squeeze(e1t)
    e2t = iris.load_cube(MESH_FILE, 'e2t').data.astype(np.float32)
    e2t = np.squeeze(e2t)
    e3t = iris.load_cube(MESH_FILE, 'e3t_0').data.astype(np.float32)
    e3t = np.squeeze(e3t)
    return e1t, e2t, e3t


def load_masks():
    basins = {}
    cubes = iris.load(REGIONS_FILE)
    for cube in cubes:
        name = cube.name()
        if name in ('nav_lat', 'nav_lon', 'Caspian_Sea'):
            continue
        cube = cube.extract(iris.Constraint(z=1))
        basins[name] = cube.data.astype(np.float32)
    return basins


def load_tos():
    tos = iris.load_cube(TOS_FILE, 'sea_surface_temperature')
    tos_data = tos.data.astype(np.float32)
    return tos_data


def load_thetao():
    thetao = iris.load_cube(THETAO_FILE, 'sea_water_potential_temperature')
    thetao_data = np.ma.filled(thetao.data, 0.0).astype(np.float32)
    return thetao_data


def compute_regmean_2D_numpy(var, basins, e1t, e2t):
    times = var.shape[0]
    regmean = {}
    for basin, mask in basins.items():
        weights = compute_weights_2D(mask, e1t, e2t)
        regmean[basin] = {}
        for t in range(times):
            regmean[basin][t] = np.average(var[t, :, :], axis=(0, 1),
                                           weights=np.squeeze(weights))
    return regmean


def compute_regmean_3D_levels(var, basins, e1t, e2t, e3t):
    times = var.shape[0]
    levs = var.shape[1]
    regmean_total = {}
    for basin, mask in basins.items():
        regmean = np.empty((times, levs))
        w = compute_weights_3D(mask, e1t, e2t, e3t)
        for t in range(times):
            for l in range(levs):
                regmean[t, l] = np.average(var[t, l, :, :], axis=(0, 1),
                                           weights=np.squeeze(w)[l, :, :])
        regmean_total[basin] = regmean
    return regmean_total


def compute_regmean_3D_cpu(var, basins, e1t, e2t, e3t):
    times = var.shape[0]
    regmean_total = {}
    for basin, mask in basins.items():
        weights = compute_weights_3D(mask, e1t, e2t, e3t)
        regmean = np.empty(times)
        for t in range(times):
            regmean[t] = np.average(var[t, :, :, :], axis=(0, 1, 2),
                                    weights=np.squeeze(weights))
        regmean_total[basin] = regmean
    return regmean_total


def compute_regmean_2D_gpu(var, basins, e1t, e2t):
    times = var.shape[0]
    lats = var.shape[1]
    lons = var.shape[2]
    mask = len(basins)

    masks = np.empty((mask, lats, lons))
    for i, basin in enumerate(basins):
        masks[i, :, :] = basins[basin]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_e1t = cuda.to_device(e1t.astype(np.float32))
    gpu_e2t = cuda.to_device(e2t.astype(np.float32))
    gpu_masks = cuda.to_device(masks)
    gpu_weighted_var = cuda.device_array((mask, times, lats*lons),
                                         dtype=np.float32)
    gpu_vector_weight = cuda.device_array((mask, lats*lons), dtype=np.float32)

    variable_weight_cuda[grid, block](gpu_var, gpu_e1t, gpu_e2t, gpu_masks,
                                      gpu_weighted_var, gpu_vector_weight)

    regmean = {}
    for m in range(mask):
        reduced_weight = sum_red_cuda(gpu_vector_weight[m, :])
        regmean[m] = {}
        for t in range(times):
            reduced_var = sum_red_cuda(gpu_weighted_var[m, t, :])
            regmean[m][t] = reduced_var / reduced_weight

    return regmean


def compute_regmean_gpu_3D(var, basins, e1t, e2t, e3t):
    times = var.shape[0]
    levs = var.shape[1]
    lats = var.shape[2]
    lons = var.shape[3]
    mask = len(basins)

    masks = np.empty((mask, lats, lons))
    for i, basin in enumerate(basins):
        masks[i, :, :] = basins[basin]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_e1t = cuda.to_device(e1t.astype(np.float32))
    gpu_e2t = cuda.to_device(e2t.astype(np.float32))
    gpu_e3t = cuda.to_device(e3t.astype(np.float32))
    gpu_masks = cuda.to_device(masks)

    gpu_weighted_var = cuda.device_array((mask, times, levs*lats*lons),
                                         dtype=np.float32)
    gpu_vector_weight = cuda.device_array((mask, levs*lats*lons),
                                          dtype=np.float32)

    variable_weight_cuda_3D[grid, block](gpu_var, gpu_e1t, gpu_e2t, gpu_e3t,
                                         gpu_masks, gpu_weighted_var,
                                         gpu_vector_weight)

    regmean = {}
    for m in range(mask):
        reduced_weight = sum_red_cuda(gpu_vector_weight[m, :])
        regmean[m] = {}
        for t in range(times):
            reduced_var = sum_red_cuda(gpu_weighted_var[m, t, :])
            regmean[m][t] = reduced_var / reduced_weight

    return regmean


def compute_regmean_lev_gpu_3D(var, basins, e1t, e2t, e3t):
    times = var.shape[0]
    levs = var.shape[1]
    lats = var.shape[2]
    lons = var.shape[3]
    mask = len(basins)

    masks = np.empty((mask, lats, lons))
    for i, basin in enumerate(basins):
        masks[i, :, :] = basins[basin]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_e1t = cuda.to_device(e1t.astype(np.float32))
    gpu_e2t = cuda.to_device(e2t.astype(np.float32))
    gpu_e3t = cuda.to_device(e3t.astype(np.float32))
    gpu_masks = cuda.to_device(masks)

    gpu_weighted_var = cuda.device_array((mask, times, levs, lats*lons),
                                         dtype=np.float32)
    gpu_vector_weight = cuda.device_array((mask, levs, lats*lons),
                                          dtype=np.float32)

    variable_weight_cuda_level_3D[grid, block](gpu_var, gpu_weights,
                                               gpu_weighted_var,
                                               gpu_vector_weight)
    del gpu_var, gpu_weights

    regmean = {}
    for m in range(mask):
        regmean[m] = {}
        for t in range(times):
            regmean[m][t] = {}
            for l in range(levs):
                reduced_weight = sum_red_cuda(gpu_vector_weight[mask, l, :])
                reduced_weighted_var = sum_red_cuda(gpu_weighted_var[mask, t,
                                                                     l, :])
                regmean[m][t][l] = reduced_weighted_var / reduced_weight
    del gpu_weighted_var, gpu_vector_weight

    return regmean


def basin_weights_2D(basins, e1t, e2t):
    weights = {}
    for basin, mask in basins.items():
        weights[basin] = compute_weights_2D(mask, e1t, e2t)
    return weights


@vectorize(['float32(float32, float32, float32)'], target='cpu')
def compute_weights_2D(mask, e1t, e2t):
    weights = mask*e1t*e2t
    return weights


@vectorize(['float32(float32, float32, float32, float32)'], target='cpu')
def compute_weights_3D(mask, e1t, e2t, e3t):
    weights = mask*e1t*e2t*e3t
    return weights


@cuda.jit
def variable_weight_cuda(var, e1t, e2t, masks, weighted_var, vector_weight):
    i, j, t = cuda.grid(3)
    basins = masks.shape[0]
    x = i+j*var.shape[2]
    if(i >= var.shape[2]):
        return
    for bas in range(basins):
        vector_weight[bas, x] = 0.0
        weighted_var[bas, t, x] = 0.0
        vector_weight[bas, x] = e1t[j, i]*e2t[j, i]*masks[bas, j, i]
        weighted_var[bas, t, x] = var[t, j, i] * vector_weight[bas, x]


@cuda.jit
def variable_weight_cuda_3D(var, e1t, e2t, e3t, masks, weighted_var,
                            vector_weight):
    i, j, t = cuda.grid(3)
    basins = masks.shape[0]
    levels = var.shape[1]
    if(i >= var.shape[3]):
        return
    for bas in range(basins):
        for lev in range(levels):
            x = i+var.shape[3]*(j+var.shape[2]*lev)
            vector_weight[bas, x] = 0.0
            vector_weight[bas, x] = (e1t[j, i] * e2t[j, i] *
                                     e3t[lev, j, i] * masks[bas, j, i])
            weighted_var[bas, t, x] = 0.0
            weighted_var[bas, t, x] = var[t, lev, j, i] * vector_weight[bas, x]


@cuda.jit
def variable_weight_cuda_level_3D(var, e1t, e2t, e3t, masks, weighted_var,
                                  vector_weight):
    i, j, t = cuda.grid(3)
    basins = masks.shape[0]
    levels = var.shape[1]
    x = i+j*var.shape[3]
    if(i >= var.shape[3]):
        return
    for bas in range(basins):
        for lev in levels:
            vector_weight[bas, lev, x] = 0.0
            vector_weight[bas, lev, x] = (e1t[j, i] * e2t[j, i] *
                                          e3t[lev, j, i] * masks[bas, j, i])
            weighted_var[bas, t, lev, x] = (var[t, lev, j, i] *
                                            vector_weight[bas, lev, x])


@vectorize(['float32(float64, float64)'], target='cuda')
def regmean_cuda(weighted_var, weights):
    regmean = weighted_var/weights
    return regmean


@cuda.reduce
def sum_red_cuda(x, y):
    return x+y


if __name__ == '__main__':
    main()
