import os
import numpy as np
import iris
import iris.cube
import iris.analysis
import warnings
import numba
import datetime
from numba import jit, njit
from numba import vectorize
from numba import guvectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver

MESH_FILE = "/esarchive/autosubmit/con_files/mesh_mask_nemo.Ec3.2_O1L75.nc"
TOS_FILE = "/esarchive/exp/ecearth/a13c/cmorfiles/CMIP/EC-Earth-Consortium"\
            "/EC-Earth3-LR/historical/r1i1p1f1/Omon/tos/gn/v20180501"\
            "/tos_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_195001-195012.nc"
THETAO_FILE = "/esarchive/exp/ecearth/a13c/cmorfiles/CMIP/EC-Earth-Consortium"\
              "/EC-Earth3-LR/historical/r1i1p1f1/Omon/thetao/gn/v20180501/"\
              "thetao_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_"\
              "195001-195012.nc"
REGIONS_FILE = "/esarchive/autosubmit/con_files/mask.regions.Ec3.2_O1L75.nc"


def main():
    e1t, e2t, e3t_0, tmask = load_mesh()
    basins = load_masks()
    tos = load_tos()
    thetao = load_thetao()
    for i in range(1):
        start = datetime.datetime.now()
        # sum_cpu = compute_regsum_2D_cpu(tos, basins, e1t, e2t)
        print('cpu {0}'.format(datetime.datetime.now()-start))
        start = datetime.datetime.now()
        # sum_gpu = compute_regsum_2D_gpu(tos, basins, e1t, e2t)
        print('gpu {0}'.format(datetime.datetime.now()-start))
        sum3D_cpu, sum_cpu = compute_regsum_3D_cpu(thetao, basins, e1t, e2t,
                                                   e3t_0, tmask, True)
        a = 0
        # sum3D_gpu = compute_regsum_3D_gpu(thetao, basins, e1t, e2t, e3t_0)
        # print(sum3D_cpu['Chukchi_Sea'][:])
        # print(sum3D_gpu[0])


def load_mesh():
    e1t = iris.load_cube(MESH_FILE, 'e1t').data.astype(np.float32)
    e1t = np.squeeze(e1t)
    e2t = iris.load_cube(MESH_FILE, 'e2t').data.astype(np.float32)
    e2t = np.squeeze(e2t)
    e3t = iris.load_cube(MESH_FILE, 'e3t_0').data.astype(np.float32)
    e3t = np.squeeze(e3t)
    tmask = iris.load_cube(MESH_FILE, 'tmask').data.astype(np.int8)
    tmask = np.squeeze(tmask)
    return e1t, e2t, e3t, tmask


def load_masks():
    basins = {}
    cubes = iris.load(REGIONS_FILE)
    for cube in cubes:
        name = cube.name()
        if name in ('nav_lat', 'nav_lon', 'Caspian_Sea'):
            continue
        if name not in ('Global_Ocean'):
            continue
        cube = cube.extract(iris.Constraint(z=1))
        basins[name] = cube.data.astype(np.float32)
    return basins


def load_tos():
    tos = iris.load_cube(TOS_FILE, 'sea_surface_temperature')
    tos_data = tos.data.astype(np.float32)
    return tos_data


def load_thetao():
    thetao = iris.load_cube(THETAO_FILE, 'sea_water_potential_temperature')
    thetao_data = np.ma.filled(thetao.data, 0.0).astype(np.float32)
    return thetao_data


def compute_regsum_2D_cpu(var, basins, e1t, e2t):
    regsum = {}
    for basin, mask in basins.items():
        weighted_var = compute_weighted_var_2D(var, mask, e1t, e2t)
        regsum[basin] = np.sum(weighted_var, axis=(1, 2))
    return regsum


def compute_regsum_3D_cpu(var, basins, e1t, e2t, e3t, tmask, save3d):
    regsum = {}
    regsum_3D = {}
    for basin, mask in basins.items():
        weighted_var = compute_weighted_var_3D(var, mask, e1t, e2t, e3t, tmask)
        regsum[basin] = np.sum(weighted_var, axis=(2, 3))
        if save3d:
            regsum_3D[basin] = np.sum(regsum[basin], axis=(1,))
        else:
            regsum_3D = None
    return regsum, regsum_3D


def compute_regsum_2D_gpu(var, basins, e1t, e2t):
    times = var.shape[0]
    lats = var.shape[1]
    lons = var.shape[2]
    mask = len(basins)

    masks = np.empty((mask, lats, lons))
    for i, basin in enumerate(basins):
        masks[i, :, :] = basins[basin]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_e1t = cuda.to_device(e1t.astype(np.float32))
    gpu_e2t = cuda.to_device(e2t.astype(np.float32))
    gpu_masks = cuda.to_device(masks)
    gpu_weighted_var = cuda.device_array((mask, times, lats*lons),
                                         dtype=np.float32)

    variable_weight_cuda[grid, block](gpu_var, gpu_e1t, gpu_e2t, gpu_masks,
                                      gpu_weighted_var)

    regsum = {}
    for m in range(mask):
        regsum[m] = {}
        for t in range(times):
            regsum[m][t] = sum_red_cuda(gpu_weighted_var[m, t, :])

    return regsum


def compute_regsum_3D_gpu(var, basins, e1t, e2t, e3t):
    times = var.shape[0]
    levs = var.shape[1]
    lats = var.shape[2]
    lons = var.shape[3]
    mask = len(basins)

    masks = np.empty((mask, lats, lons))
    for i, basin in enumerate(basins):
        masks[i, :, :] = basins[basin]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_e1t = cuda.to_device(e1t.astype(np.float32))
    gpu_e2t = cuda.to_device(e2t.astype(np.float32))
    gpu_e3t = cuda.to_device(e3t.astype(np.float32))
    gpu_masks = cuda.to_device(masks)
    gpu_weighted_var = cuda.device_array((mask, times, levs*lats*lons),
                                         dtype=np.float32)

    variable_weight_cuda_3D[grid, block](gpu_var, gpu_e1t, gpu_e2t, gpu_e3t,
                                         gpu_masks, gpu_weighted_var)

    regsum = {}
    for m in range(mask):
        regsum[m] = {}
        for t in range(times):
            regsum[m][t] = sum_red_cuda(gpu_weighted_var[m, t, :])

    return regsum

@vectorize(['float64(float32, float32, float32, float32, float32, int8)'],
           target='cpu')
def compute_weighted_var_3D(var, mask, e1t, e2t, e3t, tmask):
    weighted_var = var*mask*e1t*e2t*e3t*tmask
    return weighted_var


@vectorize(['float32(float32, float32, float32, float32)'], target='cpu')
def compute_weighted_var_2D(var, mask, e1t, e2t):
    weighted_var = var*mask*e1t*e2t
    return weighted_var


@cuda.jit
def variable_weight_cuda(var, e1t, e2t, masks, weighted_var):
    i, j, t = cuda.grid(3)
    basins = masks.shape[0]
    x = i+j*var.shape[2]
    if(i >= var.shape[2]):
        return
    for bas in range(basins):
        weighted_var[bas, t, x] = 0.0
        weighted_var[bas, t, x] = (var[t, j, i] * e1t[j, i] * e2t[j, i] *
                                   masks[bas, j, i])


@cuda.jit
def variable_weight_cuda_3D(var, e1t, e2t, e3t, masks, weighted_var):
    i, j, t = cuda.grid(3)
    basins = masks.shape[0]
    levels = var.shape[1]
    if(i >= var.shape[3]):
        return
    for bas in range(basins):
        for lev in range(levels):
            x = i+var.shape[3]*(j+var.shape[2]*lev)
            weighted_var[bas, t, x] = 0.0
            weighted_var[bas, t, x] = (var[t, j, i] * e1t[j, i] * e2t[j, i] *
                                       e3t[j, i] * masks[bas, j, i])


@cuda.reduce
def sum_red_cuda(x, y):
    return x+y


if __name__ == '__main__':
    main()
