import os

import numpy as np

import iris
import iris.cube
import iris.analysis

import warnings
import datetime

import numba
from numba import vectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver

import diagonals

__all__ = ['compute_regsum_2D', 'compute_regsum_3D']

def compute_regsum_2D(var, basins, e1, e2):
    """Function that computes the regional sum for 2D vars in the cpu.

       Computes the weights for each region and performs a weighted sum.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    e1 : float32
        Masked array containing variable e1.
    e2 : float32
        Masked array containing variable e2.

    Returns
    -------
    regmean_total: float32
        List containing regional mean for variable var.
    """
    regsum = {}
    for basin, mask in basins.items():
        weighted_var = _weigh_var_2D(var, mask, e1, e2)
        regsum[basin] = np.sum(weighted_var, axis=(1, 2))
    return regsum


def compute_regsum_3D_cpu(var, basins, e1, e2, e3, save3d):
    """Function that computes the regional sum for 3D vars in the cpu.

       Computes the weights for each region and performs a weighted sum.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    e1 : float32
        Masked array containing variable e1.
    e2 : float32
        Masked array containing variable e2.

    Returns
    -------
    regmean_total: float32
        List containing regional mean for variable var.
    """
    regsum = {}
    regsum_3D = {}
    for basin, mask in basins.items():
        weighted_var = _weigh_var_3D(var, mask, e1, e2, e3)
        regsum[basin] = np.sum(weighted_var, axis=(2, 3))
        if save3d:
            regsum_3D[basin] = np.sum(regsum[basin], axis=(1,))
        else:
            regsum_3D = None
    return regsum, regsum_3D


@vectorize(['float32(float32, float32, float32, float32)'], target='cpu')
def _weigh_var_2D(var, mask, e1, e2):
    weighted_var = var*mask*e1*e2
    return weighted_var


@vectorize(['float32(float32, float32, float32, float32, float32)'],
           target='cpu')
def _weigh_var_3D(var, mask, e1, e2, e3):
    weighted_var = var*mask*e1*e2*e3
    return weighted_var
