import os
import numpy as np
import glob
import time
import operator
import datetime

import iris
import iris.cube
import iris.analysis
import iris.coords
import iris.coord_categorisation
import iris.analysis
from iris.experimental.equalise_cubes import equalise_attributes
from iris.time import PartialDateTime

import warnings
import numba
from numba import vectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver

import diagonals

__all__ = ['compute', 'load_data']


def compute(gphit, e1t, e2t, sic_slices, basins):
    """Function that checks device and calls computing functions.

    Checks if the computations are going performed in the CPU or the GPU:

    Parameters
    ----------
    gphit : float32
        Masked array containing variable gphit.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    sic: float32
        Iris cubelist containing variable sic sliced over time.
    masks: float32
        List containing basin names and masks.

    Returns
    -------
    extn: float64
        List containing sea ice extent in the northern hemisphere at each
        timestep and basin.
    exts: float64
        List containing sea ice extent in the southern hemisphere at each
        timestep and basin.
    arean: float64
        List containing sea ice area in the northern hemisphere at each
        timestep and basin.
    areas: float64
        List containing sea ice area in the southern hemisphere at each
        timestep and basin.
    device: string
        Device where the computations are being performed (CPU or GPU)
    """
    if diagonals.CONFIG.use_gpu:
        device = 'GPU'
        extn, exts, arean, areas = _compute_sic_gpu(
            gphit, e1t, e2t, sic_slices, basins
        )
    else:
        device = 'CPU'
        extn, exts, arean, areas = _compute_sic_cpu(
            gphit, e1t, e2t, sic_slices, basins
        )

    return extn, exts, arean, areas, device


def _compute_sic_cpu(gphit, e1t, e2t, sic_slices, basins):
    """Function that computes sea ice are and extension on the CPU.

    Loops over time and basins and reduces over lat and lon:

    Parameters
    ----------
    gphit : float32
        Masked array containing variable gphit.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    sic_slices: float32
        Iris cubelist containing variable sic sliced over time.
    basins: float32
        List containing basin names and masks.

    Returns
    -------
    extn: float64
        List containing sea ice extent in the northern hemisphere at each
        timestep and basin.
    exts: float64
        List containing sea ice extent in the southern hemisphere at each
        timestep and basin.
    arean: float64
        List containing sea ice area in the northern hemisphere at each
        timestep and basin.
    areas: float64
        List containing sea ice area in the southern hemisphere at each
        timestep and basin.
    """
    extn = {}
    exts = {}
    arean = {}
    areas = {}

    for sic_data in sic_slices:
        time = sic_data.coord('time').points[0]
        extn[time] = {}
        exts[time] = {}
        arean[time] = {}
        areas[time] = {}
        sic = sic_data.data
        for basin, tmask in basins.items():
            temp = _extn_cpu(gphit, e1t, e2t, tmask, sic)
            extn[time][basin] = np.sum(temp, axis=(1, 2))

            temp = _exts_cpu(gphit, e1t, e2t, tmask, sic)
            exts[time][basin] = np.sum(temp, axis=(1, 2))

            temp = _arean_cpu(gphit, e1t, e2t, tmask, sic)
            arean[time][basin] = np.sum(temp, axis=(1, 2))

            temp = _areas_cpu(gphit, e1t, e2t, tmask, sic)
            areas[time][basin] = np.sum(temp, axis=(1, 2))

    del gphit, e1t, e2t, sic_slices, temp

    return extn, exts, arean, areas


def _compute_sic_gpu(gphit, e1t, e2t, sic_slices, basins):
    """Function that computes sea ice are and extension on the GPU.

    Loops over time and basins and reduces over lat and lon:

    Parameters
    ----------
    gphit : float32
        Masked array containing variable gphit.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    sic_slices: float32
        Iris cubelist containing variable sic sliced over time.
    basins: float32
        List containing basin names and masks.

    Returns
    -------
    extn: float64
        List containing sea ice extent in the northern hemisphere at each
        timestep and basin.
    exts: float64
        List containing sea ice extent in the southern hemisphere at each
        timestep and basin.
    arean: float64
        List containing sea ice area in the northern hemisphere at each
        timestep and basin.
    areas: float64
        List containing sea ice area in the southern hemisphere at each
        timestep and basin.
    """
    extn = {}
    exts = {}
    arean = {}
    areas = {}

    gpu_gphit = cuda.to_device(gphit.astype(np.float32))
    gpu_e1t = cuda.to_device(e1t.astype(np.float32))
    gpu_e2t = cuda.to_device(e2t.astype(np.float32))
    gpu_temp = cuda.device_array(e1t.shape[1]*e1t.shape[2], dtype=np.float32)

    block = (128, 1)
    grid_size = ((e1t.shape[2]) // block[0]) + 1
    grid = (grid_size, e1t.shape[1])

    for sic_data in sic_slices:
        time = sic_data.coord('time').points[0]
        extn[time] = {}
        exts[time] = {}
        arean[time] = {}
        areas[time] = {}
        sic = sic_data.data.astype(np.float32)
        gpu_sic = cuda.to_device(sic.astype(np.float32))
        for basin, tmask in basins.items():
            gpu_tmask = cuda.to_device(tmask.astype(np.float32))

            _extn_gpu[grid, block](gpu_gphit, gpu_e1t, gpu_e2t,
                                  gpu_tmask, gpu_sic, gpu_temp)
            extn[time][basin] = _sum_red_cuda(gpu_temp)

            _exts_gpu[grid, block](gpu_gphit, gpu_e1t, gpu_e2t,
                                   gpu_tmask, gpu_sic, gpu_temp)
            exts[time][basin] = _sum_red_cuda(gpu_temp)

            _arean_gpu[grid, block](gpu_gphit, gpu_e1t, gpu_e2t,
                                    gpu_tmask, gpu_sic, gpu_temp)
            arean[time][basin] = _sum_red_cuda(gpu_temp)

            _areas_gpu[grid, block](gpu_gphit, gpu_e1t, gpu_e2t,
                                    gpu_tmask, gpu_sic, gpu_temp)
            areas[time][basin] = _sum_red_cuda(gpu_temp)

    del gphit, e1t, e2t, sic_slices
    del gpu_gphit, gpu_e1t, gpu_e2t, gpu_tmask, gpu_sic, gpu_temp

    return extn, exts, arean, areas


def load_data(cube_gphit, cube_e1t, cube_e2t, cube_sic):
    """Function that loads data from Iris cubes.

    Loads the data from Iris cubes as float32 for cube_gphit, cube_e1t and
    cube_e2t. For cube_sic, the cube gets sliced over time:

    Parameters
    ----------
    cube_gphit : Cube
        Iris cube containing variable gphit.
    cube_e1t : Cube
        Iris cube containing variable e1t.
    cube_e2t: Cube
        Iris cube containing variable e2t.
    cube_sic: Cube
        Iris cube containing variable sic (sea_ice_area_fraction).

    Returns
    -------
    gphit : float32
        Masked array containing variable gphit.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    sic_slices: float32
        Iris cubelist containing variable sic sliced over time.
    basins: float32
        List containing basin names and masks.

    """
    gphit = cube_gphit.data.astype(np.float32)
    e1t = cube_e1t.data.astype(np.float32)
    e2t = cube_e2t.data.astype(np.float32)
    sic_slices = []
    for sic_data in cube_sic.slices_over('time'):
        sic_data.data = np.ma.filled(sic_data.data, 0.0).astype(np.float32)
        sic_slices.append(sic_data)

    return gphit, e1t, e2t, sic_slices


@vectorize(['float32(float32, float32, float32, float32, float32)'],
           target='cpu')
def _arean_cpu(gphit, e1t, e2t, tmask, sic):
    """Vectorized numba function executed on the cpu that computes the
       percentage of sea ice in the northern hemisphere for a given timestep
       and basin:

    Parameters
    ----------
    gphit : float32
        Masked array containing variable gphit.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    sic: float32
        Masked array containing variable sic for a given timestep.
    tmask: float32
        Masked array for a given basin

    Returns
    -------
    temp : float32
        Masked array to be summed over lat and lon to obtain sea ice area
    """
    if gphit > 0:
        temp = e1t * e2t * tmask * sic / 100
    return temp


@cuda.jit()
def _arean_gpu(gphit, e1t, e2t, tmask, sic, temp):
    """Numba kernel executed on the gpu that computes the percentage of sea ice
    in the northern hemisphere for a given timestep and basin:

     Parameters
     ----------
     gphit : float32
         Masked array containing variable gphit.
     e1t : float32
         Masked array containing variable e1t.
     e2t: float32
         Masked array containing variable e2t.
     sic: float32
         Masked array containing variable sic for a given timestep.
     tmask: float32
         Masked array for a given basin

     Returns
     -------
     temp : float32
         Masked array to be summed over lat and lon to obtain the sea ice area
         value. Needs to be a one dimensional array in order to be the input
         for the reduction kernel.
     """
    i, j = cuda.grid(2)
    x = i + j * e1t.shape[2]
    if(i >= e1t.shape[2]):
        return
    temp[x] = 0.0
    if(gphit[0, j, i] > 0):
        temp[x] = (e1t[0, j, i] * e2t[0, j, i] * tmask[0, j, i] * sic[j, i]
                   / 100)


@vectorize(['float32(float32, float32, float32, float32, float32)'],
           target='cpu')
def _extn_cpu(gphit, e1t, e2t, tmask, sic):
    """Vectorized numba function executed on the cpu that computes the sea ice
    extent per unit of area in the northern hemisphere for a given timestep and
    basin:

    Parameters
    ----------
    gphit : float32
        Masked array containing variable gphit.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    sic: float32
        Masked array containing variable sic for a given timestep.
    tmask: float32
        Masked array for a given basin

    Returns
    -------
    temp : float32
        Masked array to be summed over lat and lon to obtain sea ice extent
        value
    """
    if(gphit > 0 and sic > 15):
        temp = e1t * e2t * tmask
    return temp


@cuda.jit()
def _extn_gpu(gphit, e1t, e2t, tmask, sic, temp):
    """Numba kernel executed on the gpu that computes the percentage of sea ice
    extent value per unit of area in the northern hemisphere for a given
    timestep and basin:

    Parameters
    ----------
    gphit : float32
        Masked array containing variable gphit.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    sic: float32
        Masked array containing variable sic for a given timestep.
    tmask: float32
        Masked array for a given basin

    Returns
    -------
    temp : float32
        Masked array to be summed over lat and lon to obtain the sea ice extent
        value. Needs to be a one dimensional array in order to be the input
        for the reduction kernel.
    """
    i, j = cuda.grid(2)
    x = i + j * e1t.shape[2]
    if(i >= e1t.shape[2]):
        return
    temp[x] = 0.0
    if(gphit[0, j, i] > 0 and sic[j, i] > 15):
        temp[x] = e1t[0, j, i] * e2t[0, j, i] * tmask[0, j, i]


@vectorize(['float32(float32, float32, float32, float32, float32)'],
           target='cpu')
def _areas_cpu(gphit, e1t, e2t, tmask, sic):
    """Vectorized numba function executed on the cpu that computes the
       percentage of sea ice in the southern hemisphere for a given timestep
       and basin:

    Parameters
    ----------
    gphit : float32
        Masked array containing variable gphit.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    sic: float32
        Masked array containing variable sic for a given timestep.
    tmask: float32
        Masked array for a given basin

    Returns
    -------
    temp : float32
        Masked array to be summed over lat and lon to obtain sea ice area
    """
    if gphit < 0:
        temp = e1t * e2t * tmask * sic / 100
    return temp


@cuda.jit()
def _areas_gpu(gphit, e1t, e2t, tmask, sic, temp):
    """Numba kernel executed on the gpu that computes the percentage of sea ice
    in the southern hemisphere for a given timestep and basin:

    Parameters
    ----------
    gphit : float32
        Masked array containing variable gphit.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    sic: float32
        Masked array containing variable sic for a given timestep.
    tmask: float32
        Masked array for a given basin

    Returns
    -------
    temp : float32
        Masked array to be summed over lat and lon to obtain the sea ice area
        value. Needs to be a one dimensional array in order to be the input
        for the reduction kernel.
    """
    i, j = cuda.grid(2)
    x = i + j * e1t.shape[2]
    if(i >= e1t.shape[2]):
        return
    temp[x] = 0.0
    if(gphit[0, j, i] < 0):
        temp[x] = (e1t[0, j, i] * e2t[0, j, i] * tmask[0, j, i] * sic[j, i]
                   / 100)


@vectorize(['float32(float32, float32, float32, float32, float32)'],
           target='cpu')
def _exts_cpu(gphit, e1t, e2t, tmask, sic):
    """Vectorized numba function executed on the cpu that computes the sea ice
    extent per unit of area in the southern hemisphere for a given timestep and
    basin:

    Parameters
    ----------
    gphit : float32
        Masked array containing variable gphit.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    sic: float32
        Masked array containing variable sic for a given timestep.
    tmask: float32
        Masked array for a given basin

    Returns
    -------
    temp : float32
        Masked array to be summed over lat and lon to obtain sea ice extent
        value
    """
    if(gphit < 0 and sic > 15):
        temp = e1t * e2t * tmask
    return temp


@cuda.jit()
def _exts_gpu(gphit, e1t, e2t, tmask, sic, temp):
    """Numba kernel executed on the gpu that computes the percentage of sea ice
    extent value per unit of area in the southern hemisphere for a given
    timestep and basin:

    Parameters
    ----------
    gphit : float32
        Masked array containing variable gphit.
    e1t : float32
        Masked array containing variable e1t.
    e2t: float32
        Masked array containing variable e2t.
    sic: float32
        Masked array containing variable sic for a given timestep.
    tmask: float32
        Masked array for a given basin

    Returns
    -------
    temp : float32
        Masked array to be summed over lat and lon to obtain the sea ice extent
        value. Needs to be a one dimensional array in order to be the input
        for the reduction kernel.
    """
    i, j = cuda.grid(2)
    x = i + j * e1t.shape[2]
    if(i >= e1t.shape[2]):
        return
    temp[x] = 0.0
    if(gphit[0, j, i] < 0 and sic[j, i] > 15):
        temp[x] = e1t[0, j, i] * e2t[0, j, i] * tmask[0, j, i]


@vectorize(['float64(float32, float32)', 'float64(float64, float64)'],
           target='cpu')
def _sum_red(x, y):
    """Vectorized numba function executed on the cpu that performs a sum
    reduction:

    Parameters
    ----------
    x : float32
        Numpy array.
    y : float32
        Numpy array.

    Returns
    -------
    x+y : float64
        Reduction operation.
    """
    return x+y


@cuda.reduce
def _sum_red_cuda(x, y):
    """Numba kernel executed on the gpu that performs a sum reduction:

    Parameters
    ----------
    x : float32
        Numpy array.
    y : float32
        Numpy array.

    Returns
    -------
    x+y : float64
        Reduction operation.
    """
    return x+y
